from pathlib import Path

import click
import yaml  # type: ignore[import-untyped]
from ape import project
from ape.logging import logger

PLUGIN_MANIFEST_FOLDER = project.path / "ape_safe" / "manifests"

# NOTE: Exclude all non-essential contract types
VERSIONS = {
    "v1.3.0": """
compile:
  exclude:
    - accessors
    - base
    - common
    - examples
    - external
    - handler
    - interfaces
    - libraries/CreateCall.sol
    - libraries/MultiSend.sol
    - mocks
    - modules
    - proxies/IProxyCreationCallback.sol
    - test

dependencies:
  - name: openzeppelin
    github: OpenZeppelin/openzeppelin-contracts
    version: 3.4.0

solidity:
  version: 0.7.6
  import_remapping:
    - "@openzeppelin/contracts=openzeppelin/v3.4.0"
""",
    "v1.4.1": """
compile:
  exclude:
    - accessors
    - base
    - common
    - examples
    - external
    - handler
    - interfaces
    - libraries/CreateCall.sol
    - libraries/MultiSend.sol
    - libraries/SafeStorage.sol
    - libraries/SignMessageLib.sol
    - mocks
    - modules
    - proxies/IProxyCreationCallback.sol
    - test

dependencies:
  - name: openzeppelin
    github: OpenZeppelin/openzeppelin-contracts
    version: 3.4.0

solidity:
  version: 0.7.6
  import_remapping:
    - "@openzeppelin/contracts=openzeppelin/v3.4.0"
""",
}

INCLUDED_FIELDS = {
    "abi",
    "runtimeBytecode",
    "deploymentBytecode",
}


def compile_version(version: str, override: bool):
    dependency = project.dependencies.install(
        name="safe-contracts",
        github="safe-global/safe-smart-account",
        version=version,
        config_override=yaml.safe_load(VERSIONS[version]),
    )
    plugin_manifest_path = PLUGIN_MANIFEST_FOLDER / f"safe-{version}.json"
    if not override and (plugin_manifest_path).is_file():
        logger.info(f"Version '{version}' already exists, skipping.'")
        return

    if not dependency.compiled:
        logger.warning(f"Safe {version} not compiled, compiling...")
        dependency.compile()

    logger.info(f"Writing './{plugin_manifest_path.relative_to(Path.cwd())}'")
    plugin_manifest_path.write_text(
        dependency.project.manifest.model_dump_json(
            # NOTE: Create a compact representation, only what we need
            exclude={"sources": True, "compilers": True},
        )
    )


@click.command()
@click.option("--override", is_flag=True)
@click.argument(
    "versions",
    nargs=-1,
    type=click.Choice(list(VERSIONS)),
)
def cli(override, versions):
    """Build manifests for Safe protocol"""

    for version in iter(versions or VERSIONS):
        compile_version(version, override)



import click
from ape.cli import ConnectedProviderCommand, account_option, ape_cli_context
from ape.types import AddressType

from ape_safe._cli.click_ext import safe_argument


@click.group()
def modules():
    """
    Commands for handling safe modules
    """


@modules.command("list", cls=ConnectedProviderCommand)
@safe_argument
def _list(safe):
    """List all modules enabled for SAFE"""
    for module in safe.modules:
        click.echo(repr(module))


@modules.command(cls=ConnectedProviderCommand)
@safe_argument
def guard(safe):
    """Show module guard (if enabled) for SAFE"""
    if guard := safe.modules.guard:
        click.echo(f"Guard: {guard}")

    else:
        click.secho("No Module Guard set", fg="red")


@modules.command(cls=ConnectedProviderCommand)
@ape_cli_context()
@account_option()
@safe_argument
@click.option("--propose", is_flag=True, default=False)
@click.argument("module")
def enable(cli_ctx, safe, account, module, propose):
    """
    Enable MODULE for SAFE

    **WARNING**: This is a potentially destructive action and may make your safe vulnerable.
    """
    module = cli_ctx.conversion_manager.convert(module, AddressType)
    safe.modules.enable(module, submitter=account, propose=propose)


@modules.command(cls=ConnectedProviderCommand)
@ape_cli_context()
@account_option()
@safe_argument
@click.option("--propose", is_flag=True, default=False)
@click.argument("module")
def disable(cli_ctx, safe, account, module, propose):
    """
    Disable MODULE for SAFE

    **WARNING**: This is a potentially destructive action and may impact operations of your safe.
    """
    module = cli_ctx.conversion_manager.convert(module, AddressType)
    safe.modules.disable(module, submitter=account, propose=propose)



import runpy
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional, Union, cast

import click
import rich
from ape.cli import ConnectedProviderCommand, account_option
from ape.exceptions import SignatureError
from eth_typing import ChecksumAddress, Hash32
from eth_utils import humanize_hash, to_hex
from hexbytes import HexBytes

from ape_safe._cli.click_ext import (
    SafeCliContext,
    execute_option,
    safe_cli_ctx,
    safe_option,
    sender_option,
    submitter_option,
    txn_ids_argument,
)

if TYPE_CHECKING:
    from ape.api import AccountAPI

    from ape_safe.accounts import SafeAccount
    from ape_safe.client import UnexecutedTxData


@click.group()
def pending():
    """
    Commands for handling pending transactions
    """


@pending.command("list", cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@click.option("--verbose", is_flag=True)
def _list(cli_ctx, safe, verbose) -> None:
    """
    View pending transactions for a Safe
    """

    txns = list(safe.client.get_transactions(starting_nonce=safe.next_nonce, confirmed=False))
    if not txns:
        rich.print("There are no pending transactions.")
        return

    txns_by_nonce: dict[int, list[UnexecutedTxData]] = {}
    for txn in txns:
        if txn.nonce in txns_by_nonce:
            txns_by_nonce[txn.nonce].append(txn)
        else:
            txns_by_nonce[txn.nonce] = [txn]

    all_items = txns_by_nonce.items()
    total_items = len(all_items)
    max_op_len = len("rejection")
    for root_idx, (nonce, tx_list) in enumerate(all_items):
        tx_len = len(tx_list)
        for idx, tx in enumerate(tx_list):
            title = f"Transaction {nonce}"
            is_rejection = not tx.value and not tx.data and tx.to == tx.safe
            operation_name = tx.operation.name if tx.data else "transfer"
            if is_rejection:
                operation_name = "rejection"

            # Add spacing (unless verbose) so columns are aligned.
            spaces = "" if verbose else (max(0, max_op_len - len(operation_name))) * " "
            title = f"{title} {operation_name}{spaces}"
            confirmations = tx.confirmations
            rich.print(
                f"{title} "
                f"({len(confirmations)}/{safe.confirmations_required}) "
                f"safe_tx_hash={tx.safe_tx_hash}"
            )

            if verbose:
                fields = ("to", "value", "data", "base_gas", "gas_price")
                data = {}
                for field_name, value in tx.model_dump(by_alias=True, mode="json").items():
                    if field_name not in fields:
                        continue

                    if field_name in ("data",) and not value:
                        value = "0x"
                    elif not value:
                        value = "0"

                    if isinstance(value, bytes):
                        value_str = str(to_hex(HexBytes(value)))
                    else:
                        value_str = f"{value}"

                    if len(value_str) > 42:
                        value_str = f"{humanize_hash(cast(Hash32, HexBytes(value_str)))}"

                    data[field_name] = value_str

                data_str = ", ".join([f"{k}={v}" for k, v in data.items()])
                rich.print(f"\t{data_str}")
                _show_confs(tx.confirmations, extra_line=False, prefix="\t")
                if root_idx < total_items - 1 or idx < tx_len - 1:
                    click.echo()


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@click.option("--data", type=HexBytes, help="Transaction data", default=HexBytes(""))
@click.option("--gas-price", type=int, help="Transaction gas price")
@click.option("--value", type=int, help="Transaction value", default=0)
@click.option("--to", "receiver", type=ChecksumAddress, help="Transaction receiver")
@click.option("--nonce", type=int, help="Transaction nonce")
@sender_option
@click.option("--execute", help="Execute if possible after proposal", is_flag=True)
def propose(cli_ctx, ecosystem, safe, data, gas_price, value, receiver, nonce, sender, execute):
    """
    Create a new transaction
    """
    from ape.api import AccountAPI

    from ape_safe.accounts import get_signatures
    from ape_safe.utils import get_safe_tx_hash

    nonce = safe.new_nonce if nonce is None else nonce
    txn = ecosystem.create_transaction(
        value=value,
        data=data,
        gas_price=gas_price,
        nonce=nonce,
        receiver=receiver,
    )
    safe_tx = safe.create_safe_tx(txn)
    safe_tx_hash = get_safe_tx_hash(safe_tx)
    signatures = get_signatures(safe_tx, safe.local_signers)

    num_confirmations = 0
    submitter = sender if isinstance(sender, AccountAPI) else None
    if execute is None and submitter is None:
        # Check if we _can_ execute and ask the user.
        do_execute = (
            len(safe.local_signers) > 0
            and num_confirmations >= safe.confirmations_required
            and click.confirm(f"Submit transaction '{safe_tx.nonce}'")
        )
        if do_execute:
            # The user did provider a value for `--execute` however we are able to
            # So we prompt them.
            submitter = safe.select_signer(for_="submitter")

    sender = submitter if isinstance(submitter, AccountAPI) else safe.select_signer(for_="sender")
    safe.client.post_transaction(
        safe_tx,
        signatures,
        submitter=sender.address,
        contractTransactionHash=safe_tx_hash,
    )

    # Wait for new transaction to appear
    timeout = 3
    new_tx = None

    while new_tx is None and timeout > 0:
        new_tx = next(
            safe.client.get_transactions(
                starting_nonce=safe.next_nonce,
                confirmed=False,
                filter_by_ids=[safe_tx_hash],
            ),
            None,
        )
        timeout -= 1

    if new_tx:
        cli_ctx.logger.success(f"Proposed transaction '{safe_tx_hash}'.")
    else:
        cli_ctx.abort("Failed to propose transaction.")

    if execute:
        _execute(safe, new_tx, sender)


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@txn_ids_argument
@execute_option
def approve(cli_ctx: SafeCliContext, safe, txn_ids, execute):
    from ape.api import AccountAPI

    from ape_safe.utils import get_safe_tx_hash

    submitter: Optional[AccountAPI] = execute if isinstance(execute, AccountAPI) else None
    pending_transactions = list(
        safe.client.get_transactions(confirmed=False, starting_nonce=safe.next_nonce)
    )
    for txn in pending_transactions:
        # Figure out which given ID(s) we are handling.
        length_before = len(txn_ids)
        txn_ids = _filter_tx_from_ids(txn_ids, txn)
        if len(txn_ids) == length_before:
            # Not a specified txn.
            continue

        safe_tx = safe.create_safe_tx(**txn.model_dump(by_alias=True, mode="json"))
        num_confirmations = len(txn.confirmations)

        if num_confirmations < safe.confirmations_required:
            signatures_added = safe.add_signatures(safe_tx, confirmations=txn.confirmations)
            if signatures_added:
                accounts_used_str = ", ".join(list(signatures_added.keys()))
                cli_ctx.logger.success(
                    f"Signatures added to transaction '{safe_tx.nonce}' "
                    f"using accounts '{accounts_used_str}'."
                )
                num_confirmations += len(signatures_added)

        if execute is None and submitter is None:
            # Check if we _can_ execute and ask the user.
            do_execute = (
                len(safe.local_signers) > 0
                and num_confirmations >= safe.confirmations_required
                and click.confirm(f"Submit transaction '{safe_tx.nonce}'")
            )
            if do_execute:
                # The user did provider a value for `--execute` however we are able to
                # So we prompt them.
                submitter = safe.select_signer(for_="submitter")

        if submitter:
            safe_tx_hash = get_safe_tx_hash(safe_tx)
            txn.confirmations = safe.client.get_confirmations(safe_tx_hash)
            _execute(safe, txn, submitter)

    # If any txn_ids remain, they were not handled.
    if txn_ids:
        cli_ctx.abort_txns_not_found(txn_ids)


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@txn_ids_argument
# NOTE: Doesn't use --execute because we don't need BOOL values.
@submitter_option
@click.option("--nonce", help="Submitter nonce")
def execute(cli_ctx, safe, txn_ids, submitter, nonce):
    """
    Execute a transaction
    """
    pending_transactions = list(
        safe.client.get_transactions(confirmed=False, starting_nonce=safe.next_nonce)
    )

    if not submitter:
        submitter = safe.select_signer(for_="submitter")

    for txn in pending_transactions:
        # Figure out which given ID(s) we are handling.
        length_before = len(txn_ids)
        txn_ids = _filter_tx_from_ids(txn_ids, txn)
        if len(txn_ids) == length_before:
            # Not a specified txn.
            continue

        _execute(safe, txn, submitter, nonce=nonce)

    # If any txn_ids remain, they were not handled.
    if txn_ids:
        cli_ctx.abort_txns_not_found(txn_ids)


def _execute(safe: "SafeAccount", txn: "UnexecutedTxData", submitter: "AccountAPI", **tx_kwargs):
    # perf: Avoid these imports during CLI load time for `ape --help` performance.
    from ape.types import AddressType, MessageSignature

    safe_tx = safe.create_safe_tx(**txn.model_dump(mode="json", by_alias=True))
    signatures: dict[AddressType, MessageSignature] = {
        c.owner: MessageSignature.from_rsv(c.signature) for c in txn.confirmations
    }
    tx_kwargs["submitter"] = submitter
    exc_tx = safe.create_execute_transaction(safe_tx, signatures, **tx_kwargs)
    submitter.call(exc_tx)


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@txn_ids_argument
@execute_option
def reject(cli_ctx: SafeCliContext, safe, txn_ids, execute):
    """
    Reject one or more pending transactions
    """
    from ape.api import AccountAPI

    submit = execute not in (False, None)
    submitter = execute if isinstance(execute, AccountAPI) else None
    if submitter is None and submit:
        submitter = safe.select_signer(for_="submitter")

    pending_transactions = safe.client.get_transactions(
        confirmed=False, starting_nonce=safe.next_nonce
    )

    for txn in pending_transactions:
        # Figure out which given ID(s) we are handling.
        length_before = len(txn_ids)
        txn_ids = _filter_tx_from_ids(txn_ids, txn)
        if len(txn_ids) == length_before:
            # Not a specified txn.
            continue

        is_rejection = not txn.value and not txn.data and txn.to == txn.safe
        if is_rejection:
            click.echo(f"Transaction '{txn.safe_tx_hash}' already canceled!")
            continue

        elif click.confirm(f"{txn}\nCancel Transaction?"):
            try:
                safe.transfer(
                    safe,
                    0,
                    nonce=txn.nonce,
                    submit_transaction=submit,
                    submitter=submitter,
                )
            except SignatureError:
                # These are expected because of how the plugin works
                # when not submitting
                pass

            cli_ctx.logger.success(f"Canceled transaction '{txn.safe_tx_hash}'.")

    # If any txn_ids remain, they were not handled.
    if txn_ids:
        cli_ctx.abort_txns_not_found(txn_ids)


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@click.argument("txn_id")
def show_confs(cli_ctx, safe, txn_id):
    """
    Show existing confirmations
    """

    if txn_id.isnumeric():
        nonce = int(txn_id)

        # NOTE: May be more than 1 if conflicting transactions
        txns = list(
            safe.client.get_transactions(starting_nonce=nonce, ending_nonce=nonce, confirmed=False)
        )
    else:
        txns = list(
            safe.client.get_transactions(
                starting_nonce=safe.next_nonce, filter_by_ids=txn_id, confirmed=False
            )
        )

    if not txns:
        cli_ctx.abort_txns_not_found([txn_id])

    num_txns = len(txns)
    for root_idx, txn in enumerate(txns):
        header = f"Showing confirmations for transaction '{txn.nonce}'"
        operation_name = txn.operation.name if txn.data else "transfer"
        is_rejection = not txn.value and not txn.data and txn.to == txn.safe
        if is_rejection:
            header = f"{header} rejection"
        else:
            header = f"{header} {operation_name}"

        rich.print(header)
        _show_confs(txn.confirmations)
        if root_idx < num_txns - 1:
            click.echo()


def _show_confs(confs, extra_line: bool = True, prefix: Optional[str] = None):
    prefix = prefix or ""
    length = len(confs)
    for idx, conf in enumerate(confs):
        signature_str = f"[default]{humanize_hash(conf.signature)}[/default]"
        rich.print(f"{prefix}Confirmation {idx + 1} owner={conf.owner} signature='{signature_str}'")
        if extra_line and idx < length - 1:
            click.echo()


# Helper method for handling transactions in a loop.
def _filter_tx_from_ids(
    txn_ids: Sequence[Union[int, str]], txn: "UnexecutedTxData"
) -> Sequence[Union[int, str]]:
    if txn.nonce in txn_ids:
        # Filter out all transactions with the same nonce
        return [x for x in txn_ids if x != txn.nonce]

    # Handle if given nonce and hash for same txn.
    if txn.safe_tx_hash in txn_ids:
        return [x for x in txn_ids if x != txn.safe_tx_hash]

    return txn_ids


@pending.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@account_option("--submitter", prompt="Select an account to submit or propose transaction(s)")
@safe_option
def ensure(cli_ctx, ecosystem, network, submitter, safe):
    """
    Ensure Safe API queue matches `scripts/`

    This command uses scripts from `scripts/<ecosystem-name>_<network-name>_nonce*.py`
    (or `scripts/nonce*.py` if no network-specified scripts detected) and "replays" them from the
    Safe's current on-chain nonce, up to the last script available in the folder. When executed
    using a fork network, it will only perform a simulated validation of the script. When executed
    using a live network, it will propose the transaction ONLY IF the transaction at that nonce
    height does not match the expected calldata.

    To enable this, scripts under `scripts/` must be properly named, and all use
    `ape_safe.cli.propose_batch`.
    """

    if scripts_to_ensure := list(
        cli_ctx.local_project.scripts_folder.glob(
            f"{ecosystem.name}_{network.name.rstrip('-fork')}_nonce*.py"
        )
    ):
        scripts_found_str = "\n".join(
            str(fs.relative_to(cli_ctx.local_project.scripts_folder)) for fs in scripts_to_ensure
        )
        cli_ctx.logger.debug(f"Network-specific scripts found:\n{scripts_found_str}")

    elif scripts_to_ensure := list(cli_ctx.local_project.scripts_folder.glob("nonce*.py")):
        scripts_found_str = "\n".join(
            str(fs.relative_to(cli_ctx.local_project.scripts_folder)) for fs in scripts_to_ensure
        )
        cli_ctx.logger.debug(f"Scripts found:\n{scripts_found_str}")

    else:
        raise click.UsageError("No queue scripts detected under `scripts/`.")

    starting_nonce = safe.next_nonce
    if not (
        pending_scripts := {
            nonce: script_path
            for script_path in scripts_to_ensure
            if "nonce" in script_path.stem
            and (nonce := int(script_path.stem.split("nonce")[-1])) >= starting_nonce
        }
    ):
        raise click.UsageError(
            "No queue scripts detected under `scripts/` above current Safe nonce."
        )

    elif min(pending_scripts) != starting_nonce:
        raise click.UsageError(
            f"Next nonce for {safe.address} is {starting_nonce}, not {min(pending_scripts)}."
        )

    elif missing_nonces := sorted(
        set(range(starting_nonce, max(pending_scripts) + 1)) - set(pending_scripts)
    ):
        display_str = ", ".join(map(str, missing_nonces))
        raise click.UsageError(f"Missing queue scripts for nonce(s): {display_str}")

    for nonce in sorted(pending_scripts):
        script_path = pending_scripts[nonce].relative_to(cli_ctx.local_project.path)
        if not (cmd := runpy.run_path(str(script_path), run_name=script_path.stem).get("cli")):
            raise click.UsageError(f"No command `cli` detected in {script_path}.")

        cli_ctx.logger.info(
            f"Running queue script for nonce {nonce} ('{script_path}'):\n\n  {cmd.help}\n"
        )
        # NOTE: This matches signature from `ape_safe.cli:propose_batch`
        cmd.callback.__wrapped__.func(cli_ctx, network, submitter, nonce, safe)

    if not network.is_fork:
        cli_ctx.logger.success(f"Queue for {safe.address} up-to-date!")



import pytest
from ape.exceptions import ContractLogicError


@pytest.mark.parametrize("num_modules", [1, 2, 3])
def test_add_module(safe, create_module, num_modules):
    for _ in range(num_modules):
        module = create_module()

        assert module not in safe.modules
        receipt = safe.modules.enable(module)

        assert module in safe.modules

        # NOTE: SafeL2 contracts have extra event
        assert receipt.events[-2:] == [
            safe.contract.EnabledModule(module),
            safe.contract.ExecutionSuccess(),
        ]


@pytest.mark.parametrize("num_modules", [1, 2, 3])
def test_remove_module(safe, create_module, num_modules):
    for _ in range(num_modules):
        module = create_module()

        safe.modules.enable(module)
        assert module in safe.modules

    for module in safe.modules:
        receipt = safe.modules.disable(module)
        assert module not in safe.modules

        # NOTE: SafeL2 contracts have extra event
        assert receipt.events[-2:] == [
            safe.contract.DisabledModule(module),
            safe.contract.ExecutionSuccess(),
        ]


def test_module_works(safe, create_module, deployer):
    module = create_module()

    # NOTE: `handle_safe_logic_error` won't work on a generic contract call
    with pytest.raises(ContractLogicError, match="GS104"):
        module.test(safe, sender=deployer)

    safe.modules.enable(module, submitter=deployer)
    assert module in safe.modules

    module.test(safe, sender=deployer)



import pytest


@pytest.fixture(scope="module", autouse=True)
def new_multisend(VERSION):
    from ape_safe.multisend import MultiSend

    MultiSend.inject(VERSION)
    return lambda: MultiSend(version=VERSION)


def test_default_operation(safe, deployer, token, vault, mode):
    amount = 100
    token.DEBUG_mint(safe, amount, sender=deployer)

    assert vault.asset() == token
    batch = safe.create_batch()
    batch.add(token.approve, vault, amount)
    batch.add(vault.deposit, amount, safe)

    if mode == "api":
        safe_tx_id = batch.propose(safe, submitter=deployer)
        if safe.confirmations_required > 1:
            safe.add_signatures(safe_tx_id)
        receipt = safe.submit_safe_tx(safe_tx_id, submitter=deployer)

    else:
        receipt = batch(impersonate=(mode == "impersonate"), submitter=deployer)

    assert receipt.txn_hash


def test_using_multisend_from_receipts(chain, safe, deployer, token, vault):
    amount = 100
    token.DEBUG_mint(safe, amount, sender=deployer)

    assert vault.asset() == token
    batch = safe.create_batch()

    with chain.isolate():
        # HACK: Contract address must have balance to send txns directly
        safe.balance += int(1e18)

        # NOTE: Execute these transactions in an simulated context
        receipt = token.approve(vault, amount, sender=safe.address)
        assert token.allowance(safe, vault) == amount
        batch.add_from_receipt(receipt)

        receipt = vault.deposit(amount, safe, sender=safe.address)
        assert vault.balanceOf(safe) == amount
        batch.add_from_receipt(receipt)

    # NOTE: Context was reverted, so no side-effects exist
    assert token.allowance(safe, vault) == 0
    assert vault.balanceOf(safe) == 0

    batch(submitter=deployer)

    assert vault.balanceOf(safe) == amount


def test_decode_multisend(new_multisend):
    calldata = bytes.fromhex(
        "8d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000016b00527e80008d212e2891c737ba8a2768a7337d7fd200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000024f0080878000000000000000000000000584bffc5f51ccae39ad69f1c399743620e619c2b00da18f789a1d9ad33e891253660fcf1332d236b2900000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000024e74b981b000000000000000000000000584bffc5f51ccae39ad69f1c399743620e619c2b0027b5739e22ad9033bcbf192059122d163b60349d000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000247a55036500000000000000000000000000000000000000000000000000002a1b324b8f68000000000000000000000000000000000000000000"  # noqa: E501
    )
    ms = new_multisend()
    ms.add_from_calldata(calldata)
    assert ms.handler.encode_input(b"".join(ms.encoded_calls)) == calldata



import json
import os
from collections.abc import Iterator
from datetime import datetime
from functools import reduce
from typing import TYPE_CHECKING, Optional, Union, cast

from ape.types import AddressType, HexBytes, MessageSignature
from ape.utils import USER_AGENT, get_package_version
from eip712.common import SafeTxV1, SafeTxV2
from eth_utils import to_hex
from pydantic import TypeAdapter

from ape_safe.client.base import BaseSafeClient, RequestsClient
from ape_safe.client.mock import MockSafeClient
from ape_safe.client.types import (
    DelegateInfo,
    ExecutedTxData,
    OperationType,
    SafeApiTxData,
    SafeDetails,
    SafeTx,
    SafeTxConfirmation,
    SafeTxID,
    SignatureType,
    UnexecutedTxData,
    remove_null_params,
)
from ape_safe.exceptions import (
    ActionNotPerformedError,
    ClientResponseError,
    MultisigTransactionNotFoundError,
)
from ape_safe.utils import get_safe_tx_hash, order_by_signer

if TYPE_CHECKING:
    from ape.api import AccountAPI
    from requests import Response

APE_SAFE_VERSION = get_package_version(__name__)
APE_SAFE_USER_AGENT = f"Ape-Safe/{APE_SAFE_VERSION} {USER_AGENT}"
# NOTE: Origin must be a string, but can be json that contains url & name fields
ORIGIN = json.dumps(dict(url="https://apeworx.io", name="Ape Safe", ua=APE_SAFE_USER_AGENT))
assert len(ORIGIN) <= 200  # NOTE: Must be less than 200 chars

# URL for the multichain client gateway
SAFE_CLIENT_GATEWAY_URL = "https://api.safe.global/tx-service"
GATEWAY_API_KEY = os.environ.get("APE_SAFE_GATEWAY_API_KEY")
EIP3770_BLOCKCHAIN_NAMES_BY_CHAIN_ID = {
    1: "eth",  # Ethereum Mainnet
    11155111: "sep",  # Ethereum Sepolia
    10: "oeth",  # Optimism Mainnet
    42161: "arb1",  # Arbitrum One Mainnet
    56: "bnb",  # Binance Smart Chain
    146: "sonic",
    5000: "mantle",
    43114: "avax",
    1313161554: "aurora",
    8453: "base",  # Base Mainnet
    84532: "basesep",  # Base Sepolia
    42220: "celo",  # Celo Mainnet
    100: "gno",  # Gnosis Chain
    59144: "linea",  # Linea Mainnet
    137: "pol",  # Polygon
    534352: "scr",  # Scroll
    130: "unichain",  # Unichain Mainnet
    480: "wc",  # Worldchain Mainnet
    324: "zksync",  # zkSync Mainnet
    57073: "ink",  # Ink Mainnet
    800094: "berachain",  # Berachain Mainnet
}


class SafeClient(RequestsClient):
    """Client to use to communicate with Safe Client Gateway Service"""

    def __init__(
        self,
        address: AddressType,
        override_url: Optional[str] = None,
        chain_id: Optional[int] = None,
    ) -> None:
        self.address = address
        self.chain_id = chain_id

        if override_url:
            base_url = override_url
        elif chain_id:
            if chain_id not in EIP3770_BLOCKCHAIN_NAMES_BY_CHAIN_ID:
                raise ValueError(f"Chain ID {chain_id} is not a supported chain.")

            elif not GATEWAY_API_KEY:
                raise ValueError(
                    """Must provide API key via 'APE_SAFE_GATEWAY_API_KEY='.

    Get one from https://developer.safe.global if you don't have one."""
                )

            base_url = (
                f"{SAFE_CLIENT_GATEWAY_URL}/{EIP3770_BLOCKCHAIN_NAMES_BY_CHAIN_ID[chain_id]}/api"
            )
        else:
            raise ValueError("Must provide one of chain_id or override_url.")

        super().__init__(base_url)

    def _request(self, method: str, url: str, json: Optional[dict] = None, **kwargs) -> "Response":
        # NOTE: Add authorization header
        headers = kwargs.pop("headers", {})
        headers.update(dict(Authorization=f"Bearer {GATEWAY_API_KEY}"))
        return super()._request(method, url, json=json, headers=headers, **kwargs)

    @property
    def safe_details(self) -> SafeDetails:
        response = self._get(f"/safes/{self.address}")
        return SafeDetails.model_validate(response.json())

    def get_next_nonce(self) -> int:
        return self.safe_details.nonce

    def _all_transactions(self) -> Iterator[SafeApiTxData]:
        """
        Get all transactions from safe, both confirmed and unconfirmed
        """

        url = f"/safes/{self.address}/multisig-transactions"
        while url:
            response = self._get(url)
            data = response.json()

            for txn in data.get("results"):
                # NOTE: Using construct because of pydantic v2 back import validation error.
                if "isExecuted" in txn:
                    if txn["isExecuted"]:
                        yield ExecutedTxData.model_validate(txn)

                    else:
                        yield UnexecutedTxData.model_validate(txn)

                # else it is an incoming transaction

            url = data.get("next")

    def get_safe_tx(self, safe_tx_hash: SafeTxID) -> Optional[SafeApiTxData]:
        try:
            response = self._get(f"/multisig-transactions/{safe_tx_hash}", api_version="v2")
        except ClientResponseError:
            return None

        data = remove_null_params(response.json())
        return TypeAdapter(SafeApiTxData).validate_python(data)

    def get_confirmations(self, safe_tx_hash: SafeTxID) -> Iterator[SafeTxConfirmation]:
        if safe_tx := self.get_safe_tx(safe_tx_hash):
            yield from safe_tx.confirmations

    def post_transaction(
        self, safe_tx: SafeTx, signatures: dict[AddressType, MessageSignature], **kwargs
    ):
        tx_data = UnexecutedTxData.from_safe_tx(safe_tx, self.safe_details.threshold)
        signature = HexBytes(
            reduce(
                lambda raw_sig, next_sig: (
                    raw_sig
                    + (
                        next_sig.encode_rsv()
                        if isinstance(next_sig, MessageSignature)
                        else next_sig
                    )
                ),
                order_by_signer(signatures),
                b"",
            )
        )
        post_dict: dict = {
            "signature": signature.hex() if signature else None,
            "origin": ORIGIN,
        }

        for key, value in tx_data.model_dump(by_alias=True, mode="json").items():
            if isinstance(value, HexBytes):
                post_dict[key] = to_hex(value)
            elif isinstance(value, OperationType):
                post_dict[key] = int(value)
            elif isinstance(value, datetime):
                # not needed
                continue
            else:
                post_dict[key] = value

        post_dict = {**post_dict, **kwargs}

        if "signatures" in post_dict:
            # Signature handled above.
            post_dict.pop("signatures")

        url = f"/safes/{tx_data.safe}/multisig-transactions"
        response = self._post(url, json=post_dict, api_version="v2")
        return response

    def post_signatures(
        self,
        safe_tx_or_hash: Union[SafeTx, SafeTxID],
        signatures: dict[AddressType, MessageSignature],
    ):
        if isinstance(safe_tx_or_hash, (SafeTxV1, SafeTxV2)):
            safe_tx = safe_tx_or_hash
            safe_tx_hash = get_safe_tx_hash(safe_tx)
        else:
            safe_tx_hash = safe_tx_or_hash

        safe_tx_hash = cast(SafeTxID, to_hex(HexBytes(safe_tx_hash)))
        url = f"/multisig-transactions/{safe_tx_hash}/confirmations"
        signature = to_hex(
            HexBytes(b"".join([x.encode_rsv() for x in order_by_signer(signatures)]))
        )
        try:
            self._post(url, json={"signature": signature})
        except ClientResponseError as err:
            if "The requested resource was not found on this server" in err.response.text:
                raise MultisigTransactionNotFoundError(safe_tx_hash, url, err.response) from err

            raise  # The error from BaseClient we are already raising (no changes)

    def estimate_gas_cost(
        self, receiver: AddressType, value: int, data: bytes, operation: int = 0
    ) -> Optional[int]:
        url = f"/safes/{self.address}/multisig-transactions/estimations"
        request: dict = {
            "to": receiver,
            "value": value,
            "data": to_hex(HexBytes(data)),
            "operation": operation,
        }
        result = self._post(url, json=request).json()
        gas = result.get("safeTxGas")
        return int(to_hex(HexBytes(gas)), 16)

    def get_delegates(self) -> dict["AddressType", list["AddressType"]]:
        url = "/delegates"
        delegates: dict[AddressType, list[AddressType]] = {}

        while url:
            response = self._get(url, params={"safe": self.address}, api_version="v2")
            data = response.json()

            for delegate_info in map(DelegateInfo.model_validate, data.get("results", [])):
                if delegate_info.delegator not in delegates:
                    delegates[delegate_info.delegator] = [delegate_info.delegate]
                else:
                    delegates[delegate_info.delegator].append(delegate_info.delegate)

            url = data.get("next")

        return delegates

    def add_delegate(self, delegate: "AddressType", label: str, delegator: "AccountAPI"):
        msg_hash = self.create_delegate_message(delegate)

        # NOTE: This is required as Safe API uses an antiquated .signHash method
        if not (sig := delegator.sign_raw_msghash(msg_hash)):
            raise ActionNotPerformedError("Did not sign delegate approval")

        payload = {
            "safe": self.address,
            "delegate": delegate,
            "delegator": delegator.address,
            "label": label,
            "signature": sig.encode_rsv().hex(),
        }
        self._post("/delegates", json=payload, api_version="v2")

    def remove_delegate(self, delegate: "AddressType", delegator: "AccountAPI"):
        msg_hash = self.create_delegate_message(delegate)

        # NOTE: This is required as Safe API uses an antiquated .signHash method
        if not (sig := delegator.sign_raw_msghash(msg_hash)):
            raise ActionNotPerformedError("Did not sign delegate removal")

        payload = {
            "delegator": delegator.address,
            "signature": sig.encode_rsv().hex(),
        }
        self._delete(f"/delegates/{delegate}", json=payload, api_version="v2")


__all__ = [
    "ExecutedTxData",
    "BaseSafeClient",
    "MockSafeClient",
    "OperationType",
    "SafeApiTxData",
    "SafeClient",
    "SafeDetails",
    "SafeTx",
    "SafeTxConfirmation",
    "SignatureType",
    "UnexecutedTxData",
]



import itertools
import json
import os
from collections.abc import Iterable, Iterator, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Union, cast

from ape.api import AccountAPI, AccountContainerAPI, ReceiptAPI, TransactionAPI
from ape.api.networks import ForkedNetworkAPI
from ape.cli import select_account
from ape.contracts import ContractCall
from ape.exceptions import ContractNotFoundError, ProviderNotConnectedError
from ape.logging import logger
from ape.types import AddressType, MessageSignature
from ape.utils import ZERO_ADDRESS, cached_property
from ape_ethereum.proxies import ProxyInfo, ProxyType
from ape_ethereum.transactions import TransactionType
from eip712.common import SafeTxV1, SafeTxV2, create_safe_tx_def
from eth_utils import keccak, to_bytes, to_canonical_address, to_int
from ethpm_types import ContractType
from ethpm_types.abi import ABIType, MethodABI
from packaging.version import Version

from ape_safe.client.types import OperationType

from .client import BaseSafeClient, MockSafeClient, SafeClient, SafeTx, SafeTxConfirmation, SafeTxID
from .config import SafeConfig
from .exceptions import (
    ApeSafeError,
    FeatureNotAvailable,
    NoLocalSigners,
    NotASigner,
    NotEnoughSignatures,
    NoVersionDetected,
    SafeClientException,
    handle_safe_logic_error,
)
from .factory import SafeFactory
from .modules import SafeModuleManager
from .packages import MANIFESTS_BY_VERSION, PackageType
from .types import SafeCacheData
from .utils import encode_signatures, get_safe_tx_hash

if TYPE_CHECKING:
    from ape.api.address import BaseAddress
    from ape.contracts import ContractInstance

    from ape_safe.multisend import MultiSend


class SafeContainer(AccountContainerAPI):
    _accounts: dict[str, "SafeAccount"] = {}

    @property
    def config(self) -> SafeConfig:
        return cast(SafeConfig, self.config_manager["safe"])

    @property
    def _account_files(self) -> Iterator[Path]:
        account_files = list(self.data_folder.glob("*.json"))

        # NOTE: Make sure these Safes exist in our local cache
        for required_safe, safe_cache_data in self.config.require.items():
            # NOTE: If alias `required_safe` already exists, skip overwriting it
            if required_safe not in map(lambda p: p.stem, account_files):
                safe_cache_file = self.data_folder / f"{required_safe}.json"
                safe_cache_file.write_text(safe_cache_data.model_dump_json(), encoding="utf-8")
                account_files.append(safe_cache_file)

        yield from account_files

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._account_files:
            yield p.stem

    @property
    def addresses(self) -> Iterator[str]:
        for safe in self.accounts:
            yield safe.address

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for account_file in self._account_files:
            if account_file.stem in self._accounts:
                yield self._accounts[account_file.stem]

            else:
                # Cache the accounts so their local state is maintained
                # throughout the current Python session.
                acct = SafeAccount(account_file_path=account_file)
                self._accounts[account_file.stem] = acct
                yield acct

    def __len__(self) -> int:
        return len([*self._account_files])

    def __setitem__(self, alias: str, address: str):  # type: ignore[override]
        self.save_account(alias, address)

    def __delitem__(self, alias: str):
        self.delete_account(alias)

    def __iter__(self) -> Iterator["SafeAccount"]:  # type: ignore[override]
        # NOTE: We know our accounts are SafeAccounts, hence the type ignore.s
        safe_accounts = cast(Iterator["SafeAccount"], self.accounts)
        yield from safe_accounts

    def __contains__(self, item: Union[str, "SafeAccount"]) -> bool:
        if item is None:
            return False

        if isinstance(item, str):
            return item in self.aliases or item in self.addresses

        # Is account object
        for account in self.accounts:
            if account.address == item.address:
                return True

        return False

    def save_account(self, alias: str, address: str):
        """
        Save a new Safe to your ape configuration.

        Raises:
            :class:`~ape_safe.exceptions.ApeSafeError`: When the alias
              already exists.

        Args:
            alias (str): The alias to save the Safe under.
            address (str): The address of the Safe account.
        """
        chain_id = self.provider.chain_id
        account_data = {"address": address, "deployed_chain_ids": [chain_id]}
        path = self._get_path(alias)
        if path.is_file():
            raise ApeSafeError(f"Safe with alias '{alias}' already exists.")

        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(json.dumps(account_data))

    def load_account(self, alias: str) -> "SafeAccount":
        """
        Load the Safe account.

        Raises:
            :class:`~ape_safe.exceptions.ApeSafeError`: When the alias does
              not exist.

        Args:
            alias (str): The alias the Safe account is saved under.

        Returns:
            :class:`~ape_safe.accounts.SafeAccount`: The Safe account loaded.
        """
        if alias in self._accounts:
            return self._accounts[alias]

        account_path = self._get_path(alias)
        if not account_path.is_file():
            raise ApeSafeError(f"Safe with '{alias}' does not exist")

        acct = SafeAccount(account_file_path=account_path)
        self._accounts[alias] = acct
        return acct

    def delete_account(self, alias: str):
        """
        Delete the local Safe account.
        **NOTE**: If the account does not exist, nothing happens.

        Args:
            alias (str): The alias the Safe account is saved under.
        """
        self._get_path(alias).unlink(missing_ok=True)

    def _get_path(self, alias: str) -> Path:
        return self.data_folder.joinpath(f"{alias}.json")


def get_signatures(
    safe_tx: SafeTx,
    signers: Iterable[AccountAPI],
) -> dict[AddressType, MessageSignature]:
    signatures: dict[AddressType, MessageSignature] = {}
    for signer in signers:
        if signature := signer.sign_message(safe_tx):
            signatures[signer.address] = signature

    return signatures


def _safe_tx_exec_args(safe_tx: SafeTx) -> list:
    return list(safe_tx._body_["message"].values())


class SafeAccount(AccountAPI):
    account_file_path: Path  # NOTE: Cache any relevant data here
    _factory: ClassVar[SafeFactory] = SafeFactory()
    _warn_supported_version: ClassVar[bool] = True

    def __dir__(self) -> list[str]:
        return [
            "contract",
            "fallback_handler",
            "modules",
            "guard",
            "set_guard",
            "version",
            "confirmations_required",
            "signers",
            "local_signers",
            "safe_tx_def",
            "create_safe_tx",
            "propose_safe_tx",
            "submit_safe_tx",
            "client",
            "propose",
            "pending_transactions",
            "add_signatures",
            "create_execute_transaction",
            "create_batch",
            "all_delegates",
            *super().__dir__(),
        ]

    @property
    def alias(self) -> str:
        return self.account_file_path.stem

    @property
    def account_data(self) -> SafeCacheData:
        return SafeCacheData.model_validate_json(self.account_file_path.read_text(encoding="utf-8"))

    @property
    def deployed_chain_ids(self) -> list[int]:
        return self.account_data.deployed_chain_ids

    @cached_property
    def address(self) -> AddressType:
        try:
            ecosystem = self.provider.network.ecosystem
        except ProviderNotConnectedError:
            ecosystem = self.network_manager.ethereum

        return ecosystem.decode_address(self.account_data.address)

    @cached_property
    def contract(self) -> "ContractInstance":
        version = self.version
        safe_contract = PackageType.PROXY(version).at(
            self.address,
            proxy_info=ProxyInfo(
                target=self._factory.get_singleton(version=version),
                type=ProxyType.GnosisSafe,
            ),
        )
        if self.fallback_handler:
            contract_signatures = {x.signature for x in safe_contract.contract_type.abi}
            fallback_signatures = {x.signature for x in self.fallback_handler.contract_type.abi}
            if fallback_signatures < contract_signatures:
                return safe_contract  # for some reason this never gets hit

            contract_type = safe_contract.contract_type.model_dump(by_alias=True, mode="json")
            fallback_type = self.fallback_handler.contract_type.model_dump(
                by_alias=True, mode="json"
            )
            contract_type["abi"].extend(fallback_type["abi"])
            return self.chain_manager.contracts.instance_at(
                self.address, contract_type=ContractType.model_validate(contract_type)
            )

        else:
            return safe_contract

    @cached_property
    def fallback_handler(self) -> Optional["ContractInstance"]:
        slot = keccak(text="fallback_manager.handler.address")
        value = self.provider.get_storage(self.address, slot)
        address = self.network_manager.ecosystem.decode_address(value[-20:])
        return (
            self.chain_manager.contracts.instance_at(address) if address != ZERO_ADDRESS else None
        )

    @cached_property
    def modules(self) -> SafeModuleManager:
        return SafeModuleManager(self)

    @property
    def guard(self) -> Optional["ContractInstance"]:
        if self.version < Version("1.3.0"):
            raise FeatureNotAvailable("Safe Guard", self.version)

        slot = keccak(text="guard_manager.guard.address")
        value = self.provider.get_storage(self.address, slot)
        address = self.network_manager.ecosystem.decode_address(value[-20:])
        return (
            self.chain_manager.contracts.instance_at(address) if address != ZERO_ADDRESS else None
        )

    def set_guard(
        self,
        new_guard: Union[str, AddressType, "BaseAddress"] = ZERO_ADDRESS,
        **tx_args,
    ) -> ReceiptAPI:
        if self.version < Version("1.3.0"):
            raise FeatureNotAvailable("Safe Guard", self.version)

        if "sender" not in tx_args:
            tx_args["sender"] = self

        return self.contract.setGuard(new_guard, **tx_args)

    def get_client(
        self, chain_id: Optional[int] = None, override_url: Optional[str] = None
    ) -> BaseSafeClient:
        if chain_id is None:
            chain_id = self.provider.chain_id

        if override_url is None:
            env_override = os.environ.get("SAFE_TRANSACTION_SERVICE_URL")
            if env_override:
                override_url = env_override

        if chain_id == 0 or (self.provider.network.is_local and self.provider.chain_id == chain_id):
            return MockSafeClient(contract=self.contract)

        return SafeClient(address=self.address, chain_id=chain_id, override_url=override_url)

    @cached_property
    def client(self) -> BaseSafeClient:
        if (
            self.provider.network.name.endswith("-fork")
            and isinstance(self.provider.network, ForkedNetworkAPI)
            and self.provider.network.upstream_chain_id in self.deployed_chain_ids
        ):
            return self.get_client(chain_id=self.provider.network.upstream_chain_id)

        return self.get_client()

    @property
    def version(self) -> Version:
        # NOTE: We want to make a direct call, so we can use this for loading
        #       the proper verison of the Safe protocol w/ `.contract`
        VERSION_ABI = MethodABI(
            name="VERSION",
            type="function",
            stateMutability="view",
            outputs=[ABIType(type="string")],
        )

        if not isinstance(version := ContractCall(VERSION_ABI, address=self.address)(), str):
            # NOTE: If `eth_call` returns nothing, the safe is likely not on the correct network
            raise NoVersionDetected(self.address)

        elif (
            version := Version(version.replace("+L2", ""))
        ) not in MANIFESTS_BY_VERSION and self._warn_supported_version:
            supported_versions = "', 'v".join(map(str, MANIFESTS_BY_VERSION))
            logger.warning(
                f"'v{version}' is not a version supported by this library, "
                f"consider migrating to one of {supported_versions}"
            )
            self.__class__._warn_supported_version = False

        return version

    @property
    def signers(self) -> list[AddressType]:
        # NOTE: Signers are in order because of `Set`
        try:
            return self.client.safe_details.owners
        except Exception:
            return self.contract.getOwners()

    @property
    def confirmations_required(self) -> int:
        try:
            return self.client.safe_details.threshold
        except Exception:
            return self.contract.getThreshold()

    @property
    def next_nonce(self) -> int:
        """
        The next nonce for on-chain. If you have multiple transactions
        are in the queue but not published on chain, the next nonce
        refers to the earliest nonce in that queue.
        """
        try:
            return self.client.get_next_nonce()
        except Exception:
            # NOTE: `.nonce` conflicts with `AccountAPI.nonce`, so use `._view_methods_`
            return self.contract._view_methods_["nonce"]()

    @property
    def new_nonce(self):
        """
        The next unused nonce in the system. This is different
        than ``.next_nonce`` because it includes all nonces the
        transaction service is aware of and not just the next
        on-chain nonce.
        """

        # NOTE: Transaction and returned greatest nonce first.
        if latest_tx := next(self.client.get_transactions(confirmed=False), None):
            return latest_tx.nonce + 1

        # No pending transactions. Use next on-chain nonce.
        return self.next_nonce

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        # TODO: Support signing via https://eips.ethereum.org/EIPS/eip-1271 in Ape
        raise NotImplementedError("Safe accounts do not support message signing!")

    @property
    def safe_tx_def(self) -> type[SafeTx]:
        return create_safe_tx_def(
            version=str(self.version),
            contract_address=self.address,
            chain_id=self.provider.chain_id,
        )

    def create_safe_tx(self, txn: Optional[TransactionAPI] = None, **safe_tx_kwargs) -> SafeTx:
        """
        Create the Safe transaction.

        Args:
            txn (Optional[``TransactionAPI``]): The transaction
            **safe_tx_kwargs: The safe transaction specifications, such as ``operation``.

        Returns:
            :class:`~ape_safe.client.SafeTx`: The Safe Transaction to be used.
        """
        return self.safe_tx_def(  # type: ignore[call-arg]
            # NOTE: Use `txn` only for these 3 commonly-used params
            to=(
                txn.receiver
                if txn
                else self.conversion_manager.convert(
                    safe_tx_kwargs.get("to", self.address), AddressType
                )
            ),
            value=(
                txn.value
                if txn
                else self.conversion_manager.convert(safe_tx_kwargs.get("value", 0), int)
            ),
            data=(txn.data or b"") if txn else safe_tx_kwargs.get("data", b""),
            # ...then use only kwargs for the rest
            # NOTE: Use `.new_nonce` to get latest nonce in off-chain txn queue
            nonce=safe_tx_kwargs.get("nonce", self.new_nonce),
            operation=safe_tx_kwargs.get("operation", OperationType.CALL),
            safeTxGas=self.conversion_manager.convert(safe_tx_kwargs.get("safeTxGas", 0), int),
            gasPrice=self.conversion_manager.convert(safe_tx_kwargs.get("gasPrice", 0), int),
            gasToken=self.conversion_manager.convert(
                safe_tx_kwargs.get("gasToken", ZERO_ADDRESS), AddressType
            ),
            refundReceiver=self.conversion_manager.convert(
                safe_tx_kwargs.get("refundReceiver", ZERO_ADDRESS), AddressType
            ),
        )

    def create_batch(self) -> "MultiSend":
        from ape_safe.multisend import MultiSend

        return MultiSend(safe=self, version=self.version)

    def all_delegates(self) -> Iterator[AddressType]:
        for delegates in self.client.get_delegates().values():
            yield from delegates

    def propose_safe_tx(
        self,
        safe_tx: SafeTx,
        submitter: Union[AccountAPI, AddressType, str, None] = None,
        sigs_by_signer: Optional[dict[AddressType, MessageSignature]] = None,
        contractTransactionHash: Optional[SafeTxID] = None,
    ) -> SafeTxID:
        """
        Propose a safe_tx to the Safe API client
        """
        if not contractTransactionHash:
            contractTransactionHash = get_safe_tx_hash(safe_tx)

        if not sigs_by_signer:
            sigs_by_signer = {}

        if submitter is not None and not isinstance(submitter, AccountAPI):
            submitter = self.load_submitter(submitter)
            assert isinstance(submitter, AccountAPI)  # NOTE: mypy happy

        if (
            submitter is not None
            and submitter.address not in sigs_by_signer
            and len(sigs_by_signer) < self.confirmations_required
            and (submitter.address in self.signers or submitter.address in self.all_delegates())
        ):
            if sig := submitter.sign_message(safe_tx):
                sigs_by_signer[submitter.address] = sig

        # NOTE: Signatures don't have to be in order for Safe API post
        self.client.post_transaction(
            safe_tx,
            sigs_by_signer,
            sender=submitter.address if submitter else None,
            contractTransactionHash=contractTransactionHash,
        )

        return contractTransactionHash

    def propose(
        self,
        txn: Optional[TransactionAPI] = None,
        submitter: Union[AccountAPI, AddressType, str, None] = None,
        **safe_tx_kwargs,
    ) -> SafeTxID:
        """
        Propose a transaction to the Safe API client
        """
        safe_tx = self.create_safe_tx(txn=txn, **safe_tx_kwargs)
        return self.propose_safe_tx(safe_tx, submitter=submitter)

    def pending_transactions(self) -> Iterator[tuple[SafeTx, list[SafeTxConfirmation]]]:
        for executed_tx in self.client.get_transactions(confirmed=False):
            yield (
                self.create_safe_tx(**executed_tx.model_dump(mode="json", by_alias=True)),
                executed_tx.confirmations,
            )

    @property
    def local_signers(self) -> list[AccountAPI]:
        # NOTE: Is not ordered by signing order
        # TODO: Skip per user config
        # TODO: Order per user config
        accounts: Iterable[AccountAPI]
        if self.network_manager.active_provider and self.provider.network.is_dev:
            accounts = itertools.chain(
                iter(self.account_manager),
                iter(self.account_manager.test_accounts),
            )

        else:
            accounts = iter(self.account_manager)

        # Ensure the contract is available before continuing.
        # Else, return an empty list
        try:
            signers = self.signers
        except ContractNotFoundError:
            return []

        return list(filter(lambda a: a in signers, accounts))

    def _preapproved_signature(
        self, signer: Union[AddressType, "BaseAddress", str]
    ) -> MessageSignature:
        # Get the Safe-style "preapproval" signature type, which is a sentinel value used to denote
        # when a signer approved via some other method, such as `approveHash` or being `msg.sender`
        # TODO: Link documentation for this
        return MessageSignature(
            v=1,  # Approved hash (e.g. submitter is approved)
            r=(
                b"\x00" * 12
                + to_canonical_address(self.conversion_manager.convert(signer, AddressType))
            ),
            s=b"\x00" * 32,
        )

    def _impersonate_approvals(self, safe_tx: SafeTx) -> dict[AddressType, MessageSignature]:
        safe_tx_hash = self.contract.getTransactionHash(*safe_tx)

        # Bypass signature collection logic and attempt to submit by impersonation
        # NOTE: Only works for fork and local network providers that support `set_storage`
        signatures = dict()
        for signer_address in self.signers[: self.confirmations_required]:
            # NOTE: `approvedHashes` is `address => safe_tx_hash => num_confs` @ slot 8
            # TODO: Use native ape slot indexing, once available
            #       e.g. `self.contract.approvedHashes[signer_address][safe_tx_hash] = 1`
            address_bytes32 = to_bytes(hexstr=signer_address)
            address_bytes32 = b"\x00" * (32 - len(address_bytes32)) + address_bytes32
            key_hash = keccak(address_bytes32 + b"\x00" * 31 + to_bytes(8))
            slot = to_int(keccak(safe_tx_hash + key_hash))
            self.provider.set_storage(self.address, slot, b"\x00" * 31 + b"\x01")

            signatures[signer_address] = self._preapproved_signature(signer_address)

        return signatures

    @handle_safe_logic_error()
    def create_execute_transaction(
        self,
        safe_tx: SafeTx,
        signatures: Mapping[AddressType, MessageSignature],
        submitter: Union[AccountAPI, AddressType, str, None] = None,
        impersonate: bool = False,
        **txn_options,
    ) -> TransactionAPI:
        if impersonate:
            signatures = self._impersonate_approvals(safe_tx)

        elif len(signatures) < self.confirmations_required:
            raise NotEnoughSignatures(self.confirmations_required, len(signatures))

        if not isinstance(submitter, AccountAPI):
            submitter = self.load_submitter(submitter)
            assert isinstance(submitter, AccountAPI)  # NOTE: mypy happy

        exec_args = _safe_tx_exec_args(safe_tx)[:-1]  # NOTE: Skip `nonce`
        encoded_signatures = encode_signatures(signatures)

        # NOTE: executes a `ProviderAPI.prepare_transaction`, which may produce `ContractLogicError`
        return self.contract.execTransaction.as_transaction(
            *exec_args,
            encoded_signatures,
            **txn_options,
            sender=submitter,
        )

    def compute_prev_signer(self, signer: Union[str, AddressType, "BaseAddress"]) -> AddressType:
        """
        Sometimes it's handy to have "previous owner" for ownership change operations,
        this function makes it easy to calculate.
        """
        signer_address: AddressType = self.conversion_manager.convert(signer, AddressType)
        signers = self.contract.getOwners()  # NOTE: Use contract version to ensure correctness
        if signer_address not in signers:
            raise NotASigner(signer_address)

        index = signers.index(signer_address)
        if index > 0:
            return signers[index - 1]

        # NOTE: SENTINEL_OWNERS is the "previous" address to index 0
        return cast(AddressType, "0x0000000000000000000000000000000000000001")

    def load_submitter(
        self,
        submitter: Union[AddressType, str, None] = None,
    ) -> AccountAPI:
        if submitter is None:
            if len(self.local_signers) == 0:
                raise NoLocalSigners()

            return self.local_signers[0]

        elif (
            submitter_address := self.conversion_manager.convert(submitter, AddressType)
            in self.account_manager
        ):
            return self.account_manager[submitter_address]

        elif isinstance(submitter, str) and submitter in self.account_manager.aliases:
            return self.account_manager.load(submitter)

        else:
            raise ValueError(f"Cannot handle {submitter}={type(submitter)}")

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        # NOTE: Need to override `AccountAPI` behavior for balance checks
        if txn.gas_limit is None:
            txn.gas_limit = self.estimate_gas_cost(txn=txn)

        return self.provider.prepare_transaction(txn)

    def estimate_gas_cost(self, **kwargs) -> int:
        operation = kwargs.pop("operation", 0)
        txn = kwargs.pop("txn", self.as_transaction(**kwargs))
        return (
            self.client.estimate_gas_cost(
                txn.receiver or ZERO_ADDRESS, txn.value, txn.data, operation=operation
            )
            or 0
        )

    @handle_safe_logic_error()
    def call(  # type: ignore[override]
        self,
        txn: TransactionAPI,
        impersonate: bool = False,
        **call_kwargs,
    ) -> ReceiptAPI:
        # NOTE: This handles if given `submit=None'.
        default_submit = not impersonate
        call_kwargs["submit"] = (
            call_kwargs.pop("submit_transaction", call_kwargs.pop("submit", default_submit))
            or not default_submit
        )
        call_kwargs["impersonate"] = impersonate

        return super().call(txn, **call_kwargs)

    def get_api_confirmations(
        self,
        safe_tx: Union[SafeTx, SafeTxID],
    ) -> dict[AddressType, MessageSignature]:
        if isinstance(safe_tx, (SafeTxV1, SafeTxV2)):
            safe_tx_id = get_safe_tx_hash(safe_tx)

        else:
            safe_tx_id = safe_tx

        try:
            client_confirmations = self.client.get_confirmations(safe_tx_id)
        except SafeClientException as err:
            logger.error(str(err))
            return {}

        return {
            conf.owner: MessageSignature(
                r=conf.signature[:32], s=conf.signature[32:64], v=conf.signature[64]
            )
            for conf in client_confirmations
        }

    def _contract_approvals(self, safe_tx: SafeTx) -> Mapping[AddressType, MessageSignature]:
        safe_tx_hash = self.contract.getTransactionHash(*safe_tx)

        return {
            signer: self._preapproved_signature(signer)
            for signer in self.signers
            if self.contract.approvedHashes(signer, safe_tx_hash) > 0
        }

    def _all_approvals(self, safe_tx: SafeTx) -> dict[AddressType, MessageSignature]:
        approvals = self.get_api_confirmations(safe_tx)

        # NOTE: Do this last because it should take precedence
        approvals.update(self._contract_approvals(safe_tx))
        return approvals

    @handle_safe_logic_error()
    def submit_safe_tx(
        self,
        safe_tx: Union[SafeTx, SafeTxID],
        impersonate: bool = False,
        **txn_options,
    ) -> ReceiptAPI:
        """
        Submit the safe transaction using the submitter after all signatures have been collected.

        Args:
            safe_tx (``SafeTX``): The safe transaction to submit.
            submitter (Union[``AccountAPI``, ``AddressType``, str, ``None``]):
                The submitter to use for the transaction. Defaults to ``None``.

        Returns:
            ``ReceiptAPI``
        """
        if not isinstance(safe_tx, (SafeTxV1, SafeTxV2)):
            safe_tx_id = safe_tx
            safe_tx = self.client.get_safe_tx(safe_tx_id).as_safe_tx(
                version=str(self.version), chain_id=self.chain_manager.chain_id
            )

        else:
            safe_tx_id = get_safe_tx_hash(safe_tx)
        assert isinstance(safe_tx, (SafeTxV1, SafeTxV2))

        if impersonate:
            signatures = {}

        elif len(signatures := self._all_approvals(safe_tx)) < self.confirmations_required:
            signatures = self.add_signatures(safe_tx)

        txn = self.create_execute_transaction(
            safe_tx,
            signatures,
            impersonate=impersonate,
            **txn_options,
        )

        return self.provider.send_transaction(txn)

    def sign_transaction(
        self,
        txn: TransactionAPI,
        submit: bool = True,
        submitter: Union[AccountAPI, AddressType, str, None] = None,
        skip: Optional[list[Union[AccountAPI, AddressType, str]]] = None,
        signatures_required: Optional[int] = None,  # NOTE: Required if increasing threshold
        **signer_options,
    ) -> Optional[TransactionAPI]:
        """
        Sign the created safe transaction for the safe client to post.
        **NOTE** ``signatures_required`` is required if the transaction is increasting the
        threshold.

        Args:
            txn (``TransactionAPI``): The contract transaction.
            submit (bool): The option to submit the transaction. Defaults to ``True``.
            submitter (Union[``AccountAPI``, ``AddressType``, str, None]):
                Determine who is submitting the transaction. Defaults to ``None``.
            skip (Optional[list[Union[``AccountAPI, `AddressType``, str]]]):
                Allow bypassing any specified signer. Defaults to ``None``.
            signatures_required (Optional[int]):
                The amount of signers required to confirm the transaction. Defaults to ``None``.
            **signer_options: Other signer options.

        Returns:
            Optional[``TransactionAPI``]: Returns ``None`` if the transaction is successful.
        """

        if not submit and submitter:
            raise ValueError("Cannot specify a submitter if not submitting.")

        safe_tx = self.create_safe_tx(txn, **signer_options)

        # Determine who is submitting the transaction (if enough signatures are gathered)
        # NOTE: This is needed even if not submitting right now.
        submitter_account: AccountAPI = (
            self.load_submitter(submitter)
            if submitter is None or not isinstance(submitter, AccountAPI)
            else submitter
        )

        # Garner either M or M - 1 signatures, depending on if we are submitting
        # and whether the submitter is also a signer (both must be true to submit M - 1).
        # NOTE: Will skip or reorder signers based on config
        available_signers = iter(self.local_signers)

        # If number of signatures required not specified, figure out how many are needed
        if not signatures_required:
            if submit and submitter_account.address in self.signers:
                # Sender doesn't have to sign
                signatures_required = self.confirmations_required - 1
                # NOTE: Adjust signers to sign with by skipping submitter
                available_signers = filter(lambda s: s != submitter_account, available_signers)

            else:
                # Not submitting, or submitter isn't a signer, so we need all confirmations
                signatures_required = self.confirmations_required

        # Allow bypassing any specified signers (above and beyond user config)
        if skip:
            skip_addresses = [self.conversion_manager.convert(a, AddressType) for a in skip]

            def skip_signer(signer: AccountAPI):
                return signer.address not in skip_addresses

            available_signers = filter(skip_signer, available_signers)

        # Check if transaction has existing tracked signatures
        sigs_by_signer = self._all_approvals(safe_tx)
        safe_tx_hash = get_safe_tx_hash(safe_tx)

        # Attempt to fetch just enough signatures to satisfy the amount we need
        # NOTE: It is okay to have less signatures, but it never should fetch more than needed
        signers = [x for x in available_signers if x.address not in sigs_by_signer]
        if signers:
            new_signatures = get_signatures(safe_tx, signers)
            sigs_by_signer = {**sigs_by_signer, **new_signatures}

        if (
            submit
            # We have enough signatures to commit the transaction,
            # and a non-signer will submit it as their own transaction
            and len(sigs_by_signer) >= signatures_required
        ):
            # We need to encode the submitter's address for Safe to decode
            if submitter_account.address in self.signers:
                sigs_by_signer[submitter_account.address] = self._preapproved_signature(
                    submitter_account
                )

            # Inherit gas args from safe_tx, if set
            gas_args = {"gas_limit": txn.gas_limit}

            if txn.type == TransactionType.STATIC:
                gas_args["gas_price"] = txn.gas_price

            else:
                gas_args["max_fee"] = txn.max_fee
                gas_args["max_priority_fee"] = txn.max_priority_fee

            exec_transaction = self.create_execute_transaction(
                safe_tx,
                sigs_by_signer,
                **gas_args,
                submitter=submitter_account,
            )
            return submitter_account.sign_transaction(exec_transaction, **signer_options)

        elif submit:
            # NOTE: User wanted to submit transaction, but we can't, so don't publish to API
            raise NotEnoughSignatures(signatures_required, len(sigs_by_signer))

        # NOTE: Not enough signatures were obtained to publish on-chain
        logger.info(
            f"Collected {len(sigs_by_signer)}/{self.confirmations_required} signatures "
            f"for Safe {self.address}#{safe_tx.nonce}"  # TODO: put URI
        )

        # NOTE: Do not publish to API on a dev network (e.g. `mainnet-fork`) unless mocked
        if not self.provider.network.is_dev or isinstance(self.client, MockSafeClient):
            self.propose_safe_tx(
                safe_tx,
                submitter=submitter_account,
                sigs_by_signer=sigs_by_signer,
                contractTransactionHash=safe_tx_hash,
            )

        # Return None so that Ape does not try to submit the transaction.
        return None

    def add_signatures(
        self,
        safe_tx: Union[SafeTx, SafeTxID],
        confirmations: Optional[list[SafeTxConfirmation]] = None,
    ) -> dict[AddressType, MessageSignature]:
        if not isinstance(safe_tx, (SafeTxV1, SafeTxV2)):
            safe_tx_id = safe_tx
            safe_tx = self.client.get_safe_tx(safe_tx_id).as_safe_tx(
                version=str(self.version), chain_id=self.chain_manager.chain_id
            )

        else:
            safe_tx_id = get_safe_tx_hash(safe_tx)

        assert isinstance(safe_tx, (SafeTxV1, SafeTxV2))
        confirmations = confirmations or list(self.client.get_confirmations(safe_tx_id))
        if (confirmations_needed := self.confirmations_required - len(confirmations)) <= 0:
            raise ApeSafeError("Transaction has enough confirmations.")

        elif not (
            available_signers := [
                acc
                for acc in self.local_signers
                if acc.address not in set(c.owner for c in confirmations)
            ]
        ):
            raise ApeSafeError("No local signers available to sign.")

        new_signatures = {}
        for acc in available_signers:
            if signature := acc.sign_message(safe_tx):
                new_signatures[acc.address] = signature
                confirmations_needed -= 1
                if confirmations_needed <= 0:
                    break

            # else: didn't want to sign

        if new_signatures:
            # NOTE: Do not publish to API on a dev network (e.g. `mainnet-fork`) unless mocked
            if not self.provider.network.is_dev or isinstance(self.client, MockSafeClient):
                self.client.post_signatures(safe_tx_id, new_signatures)

        # NOTE: Return all signatures, both new and existing
        return {**{c.owner: c.signature for c in confirmations}, **new_signatures}

    def select_signer(self, for_: str = "submitter") -> AccountAPI:
        return select_account(prompt_message=f"Select a {for_}", key=self.local_signers)



from collections.abc import Sequence
from typing import TYPE_CHECKING, NoReturn, Optional, Union, cast

import click
from ape.cli import ApeCliContextObject, ape_cli_context
from ape.exceptions import Abort
from click import BadOptionUsage, MissingParameter

if TYPE_CHECKING:
    # perf: Keep the CLI module loading fast as possible.
    from ape.api import AccountAPI

    from ape_safe.accounts import SafeContainer


class SafeCliContext(ApeCliContextObject):
    @property
    def safes(self) -> "SafeContainer":
        # NOTE: Would only happen in local development of this plugin.
        assert "safe" in self.account_manager.containers, "Are all API methods implemented?"

        from ape_safe.accounts import SafeContainer

        safe_container = self.account_manager.containers["safe"]
        return cast(SafeContainer, safe_container)

    def abort_txns_not_found(self, txn_ids: Sequence[Union[int, str]]) -> NoReturn:
        self.abort(f"Pending transaction(s) '{', '.join([f'{x}' for x in txn_ids])}' not found.")


def safe_cli_ctx():
    return ape_cli_context(obj_type=SafeCliContext)


def _txn_ids_callback(ctx, param, value):
    value_ls = value or []
    return [int(x) if x.isnumeric() else x for x in value_ls if x]


txn_ids_argument = click.argument(
    "txn_ids", nargs=-1, callback=_txn_ids_callback, metavar="NONCE_OR_SAFE_TX_HASH(s)"
)


class CallbackFactory:
    """
    Helper class to prevent circular import and have access
    to Ape objects.
    """

    @classmethod
    def safe_callback(cls, ctx, param, value):
        from ape.utils import ManagerAccessMixin as access

        # NOTE: For some reason, the Cli CTX object is not the SafeCliCtx yet at this point.
        safes = access.account_manager.containers["safe"]
        if value is None:
            # First, check config for a default. If one is there,
            # we must use that.
            safe_config = access.config_manager.get_config("safe")
            if alias := safe_config.default_safe:
                return access.account_manager.load(alias)

            # If there is only 1 safe, just use that.
            elif len(safes) == 1:
                return next(safes.accounts)

            elif len(safes) == 0:
                raise Abort("First, add a safe account using command:\n\t`ape safe add`")

            options = ", ".join(safes.aliases)
            raise MissingParameter(message=f"Must specify one of '{options}').")

        elif value in safes.aliases:
            return access.account_manager.load(value)

        else:
            raise BadOptionUsage("--safe", f"No safe with alias '{value}'")

    @classmethod
    def submitter_callback(cls, ctx, param, val):
        if val is None:
            return None

        from ape.utils import ManagerAccessMixin as access

        if val in access.account_manager.aliases:
            return access.account_manager.load(val)

        # Account address - execute using this account.
        elif val in access.account_manager:
            return access.account_manager[val]

        # Saying "yes, execute". Use first "local signer".
        elif val.lower() in ("true", "t", "1"):
            safe = access.account_manager.load(ctx.params["alias"])
            if not safe.local_signers:
                ctx.obj.abort("Cannot use `--execute TRUE` without a local signer.")

            return safe.select_signer(for_="submitter")

        return None

    @classmethod
    def sender_callback(cls, ctx, param, val) -> Optional[Union["AccountAPI", bool]]:
        """
        Either returns the account or ``False`` meaning don't execute.
        NOTE: The handling of the `--execute` flag in the `pending` CLI
        all happens here EXCEPT if a pending tx is executable and no
        value of `--execute` was provided.
        """
        return cls._get_execute_callback(ctx, param, val, name="sender")

    @classmethod
    def execute_callback(cls, ctx, param, val) -> Optional[Union["AccountAPI", bool]]:
        """
        Either returns the account or ``False`` meaning don't execute.
        """
        return cls._get_execute_callback(ctx, param, val)

    @classmethod
    def _get_execute_callback(cls, ctx, param, val, name: str = "execute"):
        if val is None:
            # Was not given any value.
            # If it is determined in `pending` that a tx can execute,
            # the user will get prompted.
            # Avoid this by always doing `--execute false`.
            return None

        elif submitter := cls.submitter_callback(ctx, param, val):
            return submitter

        # Saying "no, do not execute", even if we could.
        elif val.lower() in ("false", "f", "0"):
            return False

        raise BadOptionUsage(
            f"--{name}",
            f"`--{name}` value '{val}` not a boolean or account identifier.",
        )


callback_factory = CallbackFactory()
safe_option = click.option("--safe", callback=callback_factory.safe_callback)
safe_argument = click.argument("safe", callback=callback_factory.safe_callback)
submitter_option = click.option(
    "--submitter",
    help="Account to execute",
    callback=callback_factory.submitter_callback,
)
sender_option = click.option("--sender", callback=callback_factory.sender_callback)
execute_option = click.option("--execute", callback=callback_factory.execute_callback)



from importlib import import_module
from typing import Any

from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from ape_safe.config import SafeConfig

    return SafeConfig


@plugins.register(plugins.AccountPlugin)
def account_types():
    from .accounts import SafeAccount, SafeContainer

    return SafeContainer, SafeAccount


def __getattr__(name: str) -> Any:
    if name == "MultiSend":
        from .multisend import MultiSend

        return MultiSend

    elif name in ("SafeAccount", "SafeContainer"):
        return getattr(import_module("ape_safe.accounts"), name)

    elif name == "SafeConfig":
        from ape_safe.config import SafeConfig

        return SafeConfig

    else:
        raise AttributeError(name)


__all__ = [
    "MultiSend",
    "SafeAccount",
    "SafeConfig",
    "SafeContainer",
]



import inspect
from typing import TYPE_CHECKING, Callable, Optional

import click
from ape.api.accounts import ImpersonatedAccount
from ape.cli import (
    ApeCliContextObject,
    ConnectedProviderCommand,
    account_option,
    ape_cli_context,
    network_option,
)

from ape_safe._cli.click_ext import safe_argument

if TYPE_CHECKING:
    from ape.api import AccountAPI, NetworkAPI

    from ape_safe.accounts import SafeAccount


def propose_from_simulation():
    """
    Create and propose a new SafeTx from transaction receipts inside a fork.

    Usage::

        @propose_from_simulation()
        # NOTE: Name of decorated function *must* be called `cli`
        # NOTE: Decorated function may have `safe` or `submitter` args in it
        def cli():
            # This entire function is executed within a fork/isolated context
            # Use normal Ape features
            my_contract = Contract("<address>")

            # Any transaction is performed as if `sender=safe.address`
            my_contract.mutableMethod(...)

            # You can make assertions that will cause your simulation to fail if tripped
            assert my_contract.viewMethod() == ...

            # You can also add conditional calls
            if my_contract.viewMethod() < some_number:
                my_contract.mutableMethod()

            # Once you are done with your transactions, the simulation will complete after exiting


        # Once the simulation is complete, the decorator will collect all receipts it finds
        # from the `safe` and collect them into a single SafeTx to propose to a public network
    """

    def decorator(
        cmd: (
            Callable[[], None]
            | Callable[["SafeAccount"], None]
            | Callable[["SafeAccount", "AccountAPI"], None]
        ),
    ) -> click.Command:
        args: dict = {}
        parameters = inspect.signature(cmd).parameters
        if "safe" in parameters:
            args["safe"] = None

        if "submitter" in parameters:
            args["submitter"] = None

        if (script_name := cmd.__module__.split(".")[-1]).startswith("nonce"):
            if not script_name[5:].isnumeric():
                raise click.UsageError(
                    f"Script 'scripts/{script_name}.py' must follow 'nonce<N>.py', "
                    "where <N> is convertible to an integer value"
                )

            script_name_nonce = int(script_name[5:])

        else:
            script_name_nonce = None  # Use `--nonce` option or lookup on-chain later

        @click.command(cls=ConnectedProviderCommand, name=cmd.__module__)
        @ape_cli_context()
        @network_option()
        @account_option(
            "--submitter",
            prompt="Select an account to submit or propose transaction(s)",
        )
        @click.option(
            "--nonce",
            type=int,
            default=None,
            help="Set value of `nonce` for txn."
            " Will parse integer value of end of script name to automatically set.",
        )
        @safe_argument
        def new_cmd(
            cli_ctx: ApeCliContextObject,
            network: "NetworkAPI",
            submitter: "AccountAPI",
            nonce: Optional[int],
            safe: "SafeAccount",
        ):
            if "safe" in args:
                args["safe"] = safe
            if "submitter" in args:
                args["submitter"] = submitter

            # TODO: Use name of script to determine nonce? If starts with `nonce<XXX>.py`
            batch = safe.create_batch()
            total_gas_used = 0

            if nonce is None:
                if script_name_nonce is not None:
                    nonce = script_name_nonce

                else:
                    nonce = safe.next_nonce

            with (
                cli_ctx.chain_manager.isolate()
                if network.is_fork
                else cli_ctx.network_manager.fork()
            ):
                with cli_ctx.account_manager.use_sender(
                    # NOTE: Use impersonated account to skip processing w/ `SafeAccount`
                    ImpersonatedAccount(raw_address=safe.address)
                ):
                    # NOTE: Make sure the safe has money for gas
                    if not safe.balance:
                        safe.balance += int(1e20)

                    cmd(*args.values())  # NOTE: 1 or more receipts should be mined from calling cmd

                cli_ctx.logger.success(
                    "Simulation complete, collecting receipts into batch to propose"
                )

                for txn in safe.history.sessional:
                    if txn.sender != safe.address:
                        raise RuntimeError("Don't execute other transactions!")

                    elif txn.failed:
                        raise RuntimeError("Transaction failed!")

                    total_gas_used += txn.gas_used
                    batch.add_from_receipt(txn)

                # NOTE: After here, we are exiting the isolation context (either fork or rollback)

            cli_ctx.logger.info(
                f"Collected {len(batch.calls)} receipts into batch (gas used: {total_gas_used})"
            )

            if len(batch.calls) > 1:
                safe_tx = batch.as_safe_tx(nonce=nonce)

            else:  # When only one transaction receipt exits, just directly call that
                cli_ctx.logger.info("Only 1 call found, calling directly instead of MultiSend")
                txn = batch.calls[0]
                safe_tx = safe.create_safe_tx(
                    to=txn["target"],
                    value=txn["value"],
                    data=txn["callData"],
                    nonce=nonce,
                )

            if network.is_fork:  # Testing, execute as a simulation (don't set nonce)
                cli_ctx.logger.info("Using fork network, dry-running SafeTx")
                safe.create_execute_transaction(safe_tx, {}, impersonate=True, submitter=submitter)

            elif not (confirmations := safe.get_api_confirmations(safe_tx)):
                # Real mainnet, propose if not already in queue
                cli_ctx.logger.info("Using public network, proposing SafeTx to Safe API")
                safe.propose_safe_tx(safe_tx, submitter=submitter)

                safe_tx_id = safe_tx._message_hash_
                cli_ctx.logger.success(f"'{safe_tx_id.to_0x_hex()}' proposed to queue")

            else:
                safe_tx_id = safe_tx._message_hash_
                cli_ctx.logger.info(
                    f"SafeTxID '{safe_tx_id.to_0x_hex()}' already in queue "
                    f"({len(confirmations)}/{safe.confirmations_required})"
                )

        new_cmd.help = cmd.__doc__
        return new_cmd

    return decorator



from ape_safe.utils import order_by_signer


def test_order_by_signer_empty():
    assert order_by_signer({}) == []


def test_order_by_signer_1_sig(accounts):
    signature = accounts[0].sign_message("hello")
    signature_map = {accounts[0].address: signature}
    assert order_by_signer(signature_map) == [signature]


def test_order_by_signer_n_sigs(accounts):
    # NOTE: acct_0 > acct_1 by integer value.
    acct_0 = accounts[0]  # 0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266
    acct_1 = accounts[1]  # 0x70997970C51812dc3A010C7d01b50e0d17dc79C8

    signature_0 = acct_0.sign_message("hello")
    signature_1 = acct_1.sign_message("hello")

    # Ensure all orders of the dict work.
    signature_map_0 = {acct_0.address: signature_0, acct_1.address: signature_1}
    signature_map_1 = {acct_1.address: signature_1, acct_0.address: signature_0}

    # We expect the signatures to be sorted in ascending order by the
    # signer's address. Here, acct_0 > acct_1 so acct_1's signature is
    # the first and acct_0's is the latter.
    expected = [signature_1, signature_0]
    act_0 = order_by_signer(signature_map_0)
    act_1 = order_by_signer(signature_map_1)
    assert act_0 == act_1 == expected



def test_add_guard(safe, guard, exec_transaction):
    assert safe.guard is None

    receipt = exec_transaction(
        safe.contract.setGuard,
        guard,
    )

    assert safe.guard == guard

    # NOTE: SafeL2 contracts have extra event
    assert receipt.events[-2:] == [
        safe.contract.ChangedGuard(guard),
        safe.contract.ExecutionSuccess(),
    ]



from typing import TYPE_CHECKING, Optional, Union

from ape.types import AddressType
from ape.utils import ZERO_ADDRESS, ManagerAccessMixin
from eth_utils import to_checksum_address
from packaging.version import Version

from ape_safe.exceptions import handle_safe_logic_error

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ape.api import AccountAPI, ReceiptAPI
    from ape.contracts import ContractInstance

    from ape_safe.accounts import SafeAccount
    from ape_safe.client import SafeTxID


class SafeModuleManager(ManagerAccessMixin):
    SENTINEL: AddressType = "0x0000000000000000000000000000000000000001"
    PAGE_SIZE: int = 100

    def __init__(self, safe: "SafeAccount"):
        self._safe = safe

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} safe={self._safe.address}"

    def __contains__(self, module: Union[str, AddressType, "ContractInstance"]) -> bool:
        if self._safe.version >= Version("1.2.0"):
            return self._safe.contract.isModuleEnabled(module)

        return module in iter(self)

    @handle_safe_logic_error()
    def enable(
        self,
        module: Union[str, AddressType, "ContractInstance"],
        submitter: Union["AccountAPI", AddressType, str, None] = None,
        propose: bool = False,
        **txn_kwargs,
    ) -> Union["ReceiptAPI", "SafeTxID"]:
        if propose:
            txn = self._safe.contract.enableModule.as_transaction(module, **txn_kwargs)
            return self._safe.propose(txn=txn, submitter=submitter)

        else:
            return self._safe.contract.enableModule(
                module,
                sender=self._safe,
                submitter=submitter,
                **txn_kwargs,
            )

    def __iter__(self) -> "Iterator[ContractInstance]":
        start_module = self.SENTINEL
        while True:
            page, start_module = self._safe.contract.getModulesPaginated(
                start_module, self.PAGE_SIZE
            )
            yield from map(self.chain_manager.contracts.instance_at, page)

            if start_module == self.SENTINEL:
                break

    def _get_previous_module(
        self, module: Union[str, AddressType, "ContractInstance"]
    ) -> "AddressType":
        prev_module = self.SENTINEL

        for next_module in self:
            if next_module == module:
                return prev_module

            prev_module = next_module

        raise AssertionError(f"Module {module} not in Safe modules for {self._safe}")

    @handle_safe_logic_error()
    def disable(
        self,
        module: Union[str, AddressType, "ContractInstance"],
        submitter: Union["AccountAPI", AddressType, str, None] = None,
        propose: bool = False,
        **txn_kwargs,
    ) -> Union["ReceiptAPI", "SafeTxID"]:
        if propose:
            txn = self._safe.contract.disableModule.as_transaction(
                self._get_previous_module(module),
                module,
                **txn_kwargs,
            )
            return self._safe.propose(txn=txn, submitter=submitter)

        else:
            return self._safe.contract.disableModule(
                self._get_previous_module(module),
                module,
                sender=self._safe,
                **txn_kwargs,
            )

    @property
    def guard(self) -> Optional["ContractInstance"]:
        if (
            module_guard_address := to_checksum_address(
                self.provider.get_storage(
                    self._safe.contract.address,
                    # NOTE: `keccak256("module_manager.module_guard.address")`
                    "0xb104e0b93118902c651344349b610029d694cfdec91c589c91ebafbcd0289947",
                )[12:]
            )
        ) == ZERO_ADDRESS:
            return None

        return self.chain_manager.contracts.instance_at(module_guard_address)

    @handle_safe_logic_error()
    def set_guard(
        self,
        guard: Union[str, AddressType, "ContractInstance"],
        submitter: Union["AccountAPI", AddressType, str, None] = None,
        propose: bool = False,
        **txn_kwargs,
    ) -> Union["ReceiptAPI", "SafeTxID"]:
        if propose:
            txn = self._safe.contract.setModuleGuard.as_transaction(guard, **txn_kwargs)
            return self._safe.propose(txn=txn, submitter=submitter)

        else:
            return self._safe.contract.setModuleGuard(guard, sender=self._safe, **txn_kwargs)

    @handle_safe_logic_error()
    def remove_guard(self, **txn_kwargs) -> Union["ReceiptAPI", "SafeTxID"]:
        return self.set_guard(ZERO_ADDRESS, **txn_kwargs)



from collections.abc import Mapping
from typing import TYPE_CHECKING

from ape.types import AddressType, HexBytes, MessageSignature
from eip712.messages import calculate_hash
from eth_pydantic_types import HexStr
from eth_utils import to_hex, to_int

if TYPE_CHECKING:
    from .client.types import SafeTx, SafeTxID


def order_by_signer(signatures: Mapping[AddressType, MessageSignature]) -> list[MessageSignature]:
    # NOTE: Must order signatures in ascending order of signer address (converted to int)
    return [signatures[signer] for signer in sorted(signatures, key=lambda a: to_int(hexstr=a))]


def encode_signatures(signatures: Mapping[AddressType, MessageSignature]) -> HexBytes:
    return HexBytes(
        b"".join(
            sig.encode_rsv() if isinstance(sig, MessageSignature) else sig
            for sig in order_by_signer(signatures)
        )
    )


def get_safe_tx_hash(safe_tx: "SafeTx") -> "SafeTxID":
    message_hash = calculate_hash(safe_tx.signable_message)
    return HexStr(to_hex(message_hash))



from typing import TYPE_CHECKING

import pytest
from ape.exceptions import SignatureError
from eth_utils import add_0x_prefix

if TYPE_CHECKING:
    from ape.contracts import ContractTransactionHandler


@pytest.fixture(params=("sign", "api", "impersonate"))
def mode(request):
    return request.param


@pytest.fixture
def exec_transaction(mode, safe):
    impersonate = mode == "impersonate"
    submit = mode != "api"

    def exec_transaction(handler: "ContractTransactionHandler", *args):
        if submit:
            return handler(
                *args,
                sender=safe,
                impersonate=impersonate,
                submit=submit,
            )

        else:
            # Attempting to execute should raise `SignatureError` and push `safe_tx` to mock client
            size = len(list(safe.client.get_transactions(confirmed=False)))
            assert size == 0

            with pytest.raises(SignatureError):
                handler(
                    *args,
                    sender=safe,
                    impersonate=impersonate,
                    submit=submit,
                )

            pending_txns = list(safe.client.get_transactions(confirmed=False))
            assert len(pending_txns) == 1
            assert len(pending_txns[0].confirmations) >= 1
            safe_tx_hash = add_0x_prefix(pending_txns[0].safe_tx_hash)

            safe_tx_data = pending_txns[0]
            safe_tx = safe.create_safe_tx(**safe_tx_data.model_dump(by_alias=True, mode="json"))

            # Ensure client confirmations works
            client_confs = list(safe.client.get_confirmations(safe_tx_hash))
            assert len(client_confs) >= 1

            # Ensure API confirmations work
            api_confs = safe.get_api_confirmations(safe_tx)
            assert len(api_confs) >= 1

            # `safe_tx` is in mock client, extract it and execute it successfully this time
            return safe.submit_safe_tx(safe_tx)

    return exec_transaction


@pytest.fixture(scope="session")
def guard(project, deployer):
    return project.Guard.deploy(sender=deployer)


@pytest.fixture(scope="session")
def create_module(project, deployer):
    def fn():
        return project.Module.deploy(sender=deployer)

    return fn



#!/usr/bin/env python
from setuptools import find_packages, setup

extras_require = {
    "test": [  # `test` GitHub Action jobs uses this
        "pytest>=6.0",  # Core testing package
        "pytest-xdist",  # multi-process runner
        "pytest-cov",  # Coverage analyzer plugin
        "hypothesis>=6.2.0,<7.0",  # Strategy-based fuzzer
        "ape-foundry>=0.8.7",  # Used as the testing provider
        "ape-solidity>=0.8",  # Needed for compiling the Safe contracts
    ],
    "lint": [
        "mypy>=1.13.0,<2",  # Static type analyzer
        "types-requests",  # Needed for mypy type shed
        "types-setuptools",  # Needed for mypy type shed
        "ruff>=0.12.0",  # Unified linter and formatter
        "mdformat>=0.7.19",  # Docs formatter and linter
        "mdformat-gfm>=0.3.5",  # Needed for formatting GitHub-flavored markdown
        "mdformat-frontmatter>=0.4.1",  # Needed for frontmatters-style headers in issue templates
        "mdformat-pyproject>=0.0.2",  # Allows configuring in pyproject.toml
    ],
    "docs": [
        "sphinx-ape>=0.1,<1",  # Use dependencies of our custom docs plugin
    ],
    "release": [  # `release` GitHub Action job uses this
        "setuptools>=75.6.0",  # Installation tool
        "wheel",  # Packaging tool
        "twine==3.8.0",  # Package upload tool
    ],
    "dev": [
        "commitizen",  # Manage commits and publishing releases
        "pre-commit",  # Ensure that linters are run prior to committing
        "pytest-watch",  # `ptw` test watcher/runner
        "IPython",  # Console for interacting
        "ipdb",  # Debugger (Must use `export PYTHONBREAKPOINT=ipdb.set_trace`)
    ],
}

# NOTE: `pip install -e .[dev]` to install package
extras_require["dev"] = (
    extras_require["test"]
    + extras_require["lint"]
    + extras_require["docs"]
    + extras_require["release"]
    + extras_require["dev"]
)

with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="ape-safe",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="""ape-safe: Gnosis Safe account plugin for Ape""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ApeWorX Ltd.",
    author_email="admin@apeworx.io",
    url="https://github.com/ApeWorX/ape-safe",
    include_package_data=True,
    install_requires=[
        "click>=8.1.7,<9",
        "eip712>=0.2.11,<3",
        "eth-ape>=0.8.40,<0.9",
        "eth-utils>=2.1.0,<6",
        "pydantic>=2.10.2,<3",
        "requests>=2.31.0,<3",
        "requests-ratelimiter>=0.7,<1",
    ],
    entry_points={
        "ape_cli_subcommands": [
            "ape_safe=ape_safe._cli:cli",
        ],
    },
    python_requires=">=3.9,<4",
    extras_require=extras_require,
    py_modules=["ape_safe"],
    license="Apache-2.0",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"ape_safe": ["py.typed", "manifests/*"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)
